np.singleindex.ejl <-
  function(xdat = stop("training data xdat missing"),
           ydat = stop("training data ydat missing"),
           gdat = stop("training data gdat missing"),
           wdat = stop("training weights wdat missing"),
           bws, bandwidth.compute = TRUE, nmulti, random.seed = 42,
           optim.method = c("Nelder-Mead", "BFGS", "CG"),
           optim.maxattempts = 10,
           optim.reltol = sqrt(.Machine$double.eps),
           optim.abstol = .Machine$double.eps,
           optim.maxit = 500,
           ...){

    set.seed(random.seed)
    xdat = toFrame(xdat)

    if (missing(nmulti)){
      nmulti <- min(5,ncol(xdat))
    }
    
    if (ncol(xdat) < 2) {
      if (coarseclass(xdat[,1]) != "numeric")
        stop("xdat must contain at least one continuous variable")
      
      warning(paste("xdat has one dimension. Using a single index model to reduce",
                    "dimensionality is unnecessary."))
    }

    optim.method <- match.arg(optim.method)
    
    ## catch and destroy NA's
    goodrows = 1:dim(xdat)[1]
    rows.omit = attr(na.omit(data.frame(xdat,ydat)), "na.action")
    goodrows[rows.omit] = 0

    if (all(goodrows==0))
      stop("Data has no rows without NAs")

    xdat = xdat[goodrows,,drop = FALSE]
    ydat = ydat[goodrows]
    gdat = gdat[goodrows]
    
    ## convert to numeric
    if (is.factor(ydat))
      ydat <- dlev(ydat)[as.integer(ydat)]
    else
      ydat <- as.double(ydat)
    
    xdat = toMatrix(xdat)

      ichimuraMaxPenalty <- 10*mean(ydat^2)
      ichimuraFloor <- sqrt(.Machine$double.eps)
      
      ichimura <- function(param) {

        ##Define the leave-one-out objective function, sum (y - \hat
        ## G(X\hat\beta))^2. We let beta denote beta_2...beta_k (first k-1
        ## parameters in `param') and then let h denote the kth column.
        if (ncol(xdat) == 1)
          beta = numeric(0)
        else 
          beta <- param[1:(ncol(xdat)-1)]
        
        h <- param[ncol(xdat)]

        ## Next we define the sum of squared leave-one-out residuals

        sum.squares.leave.one.out <- function(xdat,ydat,gdat,beta,h) {

          ## Normalize beta_1 = 1 hence multiply X by c(1,beta)
          
          index <- as.matrix(xdat) %*% c(1,beta)

          ## One call to npksum to avoid repeated computation of the
          ## product kernel (the expensive part)

          W <- as.matrix(data.frame(ydat,1))

          tww <- npksum(txdat=cbind(index,gdat),
                        tydat=W,
                        weights=W,
                        leave.one.out=TRUE,
                        bandwidth.divide=TRUE,bws=c(h,h))$ksum

          denom <- tww[2,2,]
          denom[which(denom == 0.0)] <- ichimuraFloor
          fit.loo <- tww[1,2,]/denom
          
          wdat[which(wdat == 0.0)] <- ichimuraFloor
          # 02/27/2010: before it was (wdat^-1)*...
          t.ret <- mean(wdat*(ydat-fit.loo)^2)
          return(t.ret)

        }

        ## For the objective function, we require a positive bandwidth, so
        ## return an infinite penalty for negative h

        if(all(h > 0)) {
          return(sum.squares.leave.one.out(xdat,ydat,gdat,beta,h))
        } else {
          return(ichimuraMaxPenalty)
        }

      }
      
         ## Now we implement multistarting

         fval.min <- .Machine$double.xmax
         numimp <- 0
         fval.value <- numeric(nmulti)

         console <- newLineConsole()
         
         optim.fn <- ichimura
         optim.control <- c(abstol=optim.abstol,
                           reltol=optim.reltol,
                           maxit=optim.maxit)
        for(i in 1:nmulti) {

        console <- printPush(paste(sep="", "Multistart ", i, " of ", nmulti, "..."), console)
        ##cv.console <- newLineConsole(console)
        
        n <- nrow(xdat)

        ## We use the nlm command to minimize the objective function using
        ## starting values. Note that since we normalize beta_1=1 here beta
        ## is the k-1 vector containing beta_2...beta_k

        if(i == 1) {

          ## Initial values taken from OLS fit with a constant used for
          ## multistart 1
          ols.fit <- lm(ydat~xdat,x=TRUE)
          fit <- fitted(ols.fit)
          
          if (ncol(xdat) != 1){
            if (setequal(bws$beta[2:ncol(xdat)],c(0)))
              beta <- coef(ols.fit)[3:ncol(ols.fit$x)]
            else
              beta = bws$beta[2:ncol(xdat)]
          } else { beta = numeric(0) }

          if (bws$bw == 0)
            if(IQR(fit) > 0) {
              h <- (4/3)^0.2*min(min(sd(fit),sd(gdat)),min(IQR(fit),IQR(gdat)/(qnorm(.25,lower.tail=F)*2)))*n^(-1/5)
            } else {
             h <- (4/3)^0.2*min(sd(fit),sd(gdat))*n^(-1/5)
            }
          else
            h <- c(bws$bw)
        } else {
          ## Random initialization used for remaining multistarts

          beta.length <- length(coef(ols.fit)[3:ncol(ols.fit$x)])
          beta <- runif(beta.length,min=0.5,max=1.5)*coef(ols.fit)[3:ncol(ols.fit$x)]
          if(IQR(fit) > 0) {
            h <- runif(1,min=0.5,max=1.5)*min(min(sd(fit),sd(gdat)),min(IQR(fit),IQR(gdat)/(qnorm(.25,lower.tail=F)*2)))*n^(-1/5)
          } else {
            h <- runif(1,min=0.5,max=1.5)*min(sd(fit),sd(gdat))*n^(-1/5)
          }
        }
        
        suppressWarnings(optim.return <- optim(c(beta,h),fn=optim.fn,gr=NULL,method=optim.method,control=optim.control))
        attempts <- 0
        while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
          attempts <- attempts + 1
          beta.length <- length(coef(ols.fit)[3:ncol(ols.fit$x)])
          beta <- runif(beta.length,min=0.5,max=1.5)*coef(ols.fit)[3:ncol(ols.fit$x)]
          if(IQR(fit) > 0) {
            h <- runif(1,min=0.5,max=1.5)*min(min(sd(fit),sd(gdat)),min(IQR(fit),IQR(gdat)/(qnorm(.25,lower.tail=F)*2)))*n^(-1/5)
          } else {
            h <- runif(1,min=0.5,max=1.5)*min(sd(fit),sd(gdat))*n^(-1/5)
          }
          optim.control <- lapply(optim.control,'*',10.0)
          suppressWarnings(optim.return <- optim(c(beta,h),fn=optim.fn,gr=NULL,method=optim.method,control=optim.control))
        }
        

        if(optim.return$convergence != 0)
          stop(paste("optim failed to converge after optim.maxattempts = ", optim.maxattempts, " iterations."))


        fval.value[i] <- optim.return$value       
        if(optim.return$value < fval.min) {        
          param <- optim.return$par
          fval.min <- optim.return$value
          numimp <- numimp + 1
          best <- i
        }

        
        if(i != nmulti)
          console <- printPop(console)
      }
      console <- printClear(console)

      bws$beta <- if(ncol(xdat) == 1) 1.0
      else c(1,param[1:(ncol(xdat)-1)])
      bws$bw <- param[ncol(xdat)]
      bws$fval <- fval.min
      bws$ifval <- best
      bws$numimp <- numimp
      bws$fval.vector <- fval.value
      return(bws)
}
